/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: transfer.c
 */

#include <frontend.h>
#include <gtk/gtk.h>

#include "support.h"
#include "transfer.h"
#include "thing.h"
#include "readable.h"
#include "logging.h"
#include "help.h"

/*
 *
 *   gchar* get_container_name (object_handle_t)
 *
 *   Description:
 *      This routine returns a dynamically allocated
 *      string containing the name of the container
 *      corresponding to the handle given.
 *
 *   Entry:
 *      handle - object handle
 *
 *   Exit:
 *      Container name string which should be freed
 *      with g_free or NULL if the name could not be
 *      obtained.
 *
 */
gchar *get_container_name(object_handle_t handle)
{
	gchar *name = NULL;

	if (handle) {
		gint rc;
		handle_object_info_t *info;

		rc = evms_get_info(handle, &info);

		if (rc != SUCCESS) {
			log_error("%s: evms_get_info() returned error %d.\n", __FUNCTION__, rc);
		} else {
			name = g_strdup(info->info.container.name);
			evms_free(info);
		}
	}

	return name;
}

/*
 *
 *   void on_remove_thing_from_container_button_clicked (GtkButton *, gpointer)
 *
 *   Description:
 *      This routine initiates removing a storage object from its
 *      consuming container.
 *
 *   Entry:
 *      button    - address of the GtkButton widget
 *      user_data - object handle
 *
 *   Exit:
 *      See description.
 *
 */
void on_remove_thing_from_container_button_clicked(GtkButton * button, gpointer user_data)
{
	GtkCList *clist;
	object_handle_t handle;

	clist = GTK_CLIST(lookup_widget(GTK_WIDGET(button), "selection_window_clist"));

	handle = GPOINTER_TO_UINT(get_single_select_current_row_data(clist));

	if (handle != 0) {
		gint rc;
		handle_object_info_t * info;

		rc = evms_get_info(handle,&info);
		if (rc == SUCCESS) {
			if (info->info.object.consuming_container != 0) {
				gchar ha[sizeof(handle_array_t) + sizeof(object_handle_t)];
				handle_array_t *objects = (handle_array_t *) ha;

				objects->count = 1;
				objects->handle[0] = handle;
				rc = evms_shrink(info->info.object.consuming_container, objects, NULL);

			} else {
				log_error("%s: object %s does not belong to a container.\n", __FUNCTION__, info->info.object.name);
				rc = EINVAL;
			}

			evms_free(info);
		}

		display_selection_window_results(GTK_WIDGET(button), rc,
						 _("An error was encountered attempting to remove the storage object."),
						 _("The storage object was successfully removed."));
	}
}

/*
 *
 *   void on_remove_thing_from_container_menu_item_activate (GtkMenuItem *, gpointer)
 *
 *   Description:
 *      This routine will display the thing we intend to remove from
 *      a certain container and allow the user to confirm the removal
 *      operation.
 *
 *   Entry:
 *      menuitem  - the menuitem that initiated the action
 *      user_data - object handle
 *
 *   Exit:
 *      See description.
 *
 */
void on_remove_thing_from_container_menu_item_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	gint rc;
	object_handle_t handle = GPOINTER_TO_UINT(user_data);
	handle_object_info_t *info;

	rc = evms_get_info(handle, &info);

	if (rc != SUCCESS) {
		log_error("%s: evms_get_info() returned error %d.\n", __FUNCTION__, rc);
	} else {
		gchar *window_title;
		gchar *help_text;
		gchar *container_name;
		GtkWidget *window;
		object_type_t type;

		container_name = get_container_name(info->info.object.consuming_container);

		if (container_name) {
			window_title = g_strdup_printf(_("Remove %s from %s"),
						       info->info.object.name, container_name);
			g_free(container_name);
		} else
			window_title = g_strdup_printf(_("Remove %s from unknown container"),
						       info->info.object.name);

		if (evms_get_handle_object_type(handle, &type) == SUCCESS) {
			switch (type) {
			case REGION:
				help_text = get_remove_region_from_container_help_text();
				break;

			case SEGMENT:
				help_text = get_remove_segment_from_container_help_text();
				break;

			case DISK:
				help_text = get_remove_disk_from_container_help_text();
				break;

			default:
				help_text = NULL;
				break;
			}
		} else {
			help_text = NULL;
		}

		window = create_standard_selection_window(window_title,
							  _("Remove"),
							  help_text,
							  add_thing_as_selected_list_item,
							  on_remove_thing_from_container_button_clicked,
							  NULL, NULL, NULL, NULL, user_data);

		gtk_widget_show(window);

		evms_free(info);
		g_free(window_title);
	}
}

/*
 *
 *   void add_consumed_object_to_selection_list (GtkCList *, object_handle_t, gboolean)
 *
 *   Description:
 *      This routine appends a row to a GtkCList with information
 *      on the given object consumed by a container.
 *
 *   Entry:
 *      clist       - address of the selections GtkCList widget
 *      object      - the handle to the consumed object
 *      is_selected - whether to mark this item as selected
 *
 *   Exit:
 *      A new row corresponding to the consumed object is added to the
 *      clist.
 *
 */
void add_consumed_object_to_selection_list(GtkCList * clist, object_handle_t object,
					   gboolean is_selected)
{
	gint row;

	row = add_thing_to_selection_list(clist, object, is_selected);

	if (row != -1) {
		gint rc;
		handle_object_info_t *info;

		rc = evms_get_info(object, &info);

		if (rc != SUCCESS) {
			log_error("%s: evms_get_info() returned error %d.\n", __FUNCTION__, rc);
		} else {
			gchar *container_name;

			container_name = get_container_name(info->info.object.consuming_container);
			gtk_clist_set_text(clist, row, SL_MINMAX_SIZE_COLUMN, container_name);
			g_free(container_name);
			evms_free(info);
		}
	}
}

/*
 *
 *   gboolean can_remove_from_container (object_handle_t handle)
 *
 *   Description:
 *      This routine detremines if an object can be removed from a
 *      container by starting a shrink task on the container and
 *      seeing if the object appears in the acceptable objects list.
 *
 *   Entry:
 *      handle    - handle fo object to be removed
 *
 *   Exit:
 *      TRUE if the object can be removed from the contaner,
 *      else FALSE.
 *
 */
gboolean can_remove_from_container(object_handle_t handle)
{
	gboolean result = FALSE;
	gint rc;
	handle_object_info_t * info;

	rc = evms_get_info(handle, &info);
	if (rc != SUCCESS) {
		return result;
	}

	if (info->info.disk.consuming_container != 0) {
		task_handle_t task;
		rc = evms_create_task(info->info.disk.consuming_container,EVMS_Task_Shrink,&task);
		if (rc == SUCCESS) {
			handle_array_t *objects;
			rc == evms_get_acceptable_objects(task, &objects);
			if (rc == SUCCESS) {
				gint i;

				for (i = 0; i < objects->count; i++) {
					if (objects->handle[i] == handle) {
						result = TRUE;
						break;
					}
				}
				evms_free(objects);
			}
			evms_destroy_task(task);
		}
	}

	return result;
}

/*
 *
 *   void on_remove_thing_from_container_clist_realize (GtkWidget *, gpointer)
 *
 *   Description:
 *      This routine populates the given GtkCList with the list
 *      of storage objects that can be removed from their consuming
 *      container.
 *
 *   Entry:
 *      widget    - address of the selections GtkCList widget
 *      user_data - not used
 *
 *   Exit:
 *      Selection list populated with storage objects in the handle
 *      array that can removed from their consuming container
 *
 */
void on_remove_thing_from_container_clist_realize(GtkWidget * widget, gpointer user_data)
{
	gint rc = 0;
	GtkCList *clist = GTK_CLIST(widget);
	handle_array_t *things;

	/*
	 * Use the normally hidden min/max size column for the name of the consuming
	 * container.
	 */
	gtk_clist_set_column_visibility(clist, SL_MINMAX_SIZE_COLUMN, TRUE);
	gtk_clist_set_column_auto_resize(clist, SL_MINMAX_SIZE_COLUMN, TRUE);

	set_selection_window_clist_column_titles(clist, _("Size"), _("Storage Object"),
						 _("Consuming Container"));

	rc = evms_get_object_list(0, DATA_TYPE, 0, 0, 0, &things);

	if (rc != SUCCESS) {
		log_error("%s: evms_get_object_list() returned error code %d.\n", __FUNCTION__, rc);
	} else {
		guint i;
		gboolean is_selected = (things->count == 1);

		for (i = 0; i < things->count; i++) {
			if (can_remove_from_container(things->handle[i]))
				add_consumed_object_to_selection_list(clist, things->handle[i],
								      is_selected);
		}

		if (clist->rows == 1)
			gtk_clist_select_row(clist, 0, 0);

		evms_free(things);
	}
}
